﻿# Twic2Scid.ps1 1.0 unzips the downloaded PGN then backups the Scid base of reference and imports the PGN
Set-StrictMode -Version 2.0 # Freeware without any garanties (*). Powered by echiquierBriochin.fr
# Setup: 1. Create the "twic" subfolder of the Scid database folder "bin\bases\"
#        2. Download this script in "bin\bases\twic\" from echiquierBriochin.fr
#        3. Click the start button of Windows 10. Enter letter "i" like ISE then run the GUI of Powershell ISE 
#           or right click Win10 start button > contextual menu "Windows Powershell"
#        4. In Windows Powershell ISE or in NotePad++, configure the main variables of the script until line 76
#        5. In the dark blue console of Windows Powershell (ISE), enter:
#           .\Twic2Scid.ps1
#           Or click the green triangle "Run the script" (F5) in the toolbar button of Powershell ISE

# local scid on the hard disk
$pathScid = "C:\Chess\Scid\Prog\bin" # should exist. Indicate the path of your installation of Scid
#                          ^we separate the folder "Prog" (standing for program) from the "Setup" folder
# We do NOT install Scid vs. PC in "C:\Program Files" that requires admin rights for any modifications of Tcl/Tk GUI
$dirRef = Join-Path -Path $pathScid -ChildPath "bases" # Scid database folder created by Scid setup
$dirTwic = Join-Path -Path $dirRef -ChildPath "twic" # should exist. Folder of this script
$ScidTheWeekInChess = "twic" # Name of the weekly Scid base without extension
$ScidFullTwic =  Join-Path -Path $dirTwic -ChildPath $ScidTheWeekInChess

# remote PGN on the internet identified by its twic number in string format:
$twicNbr = "1316" # The week in chess last number on January 29, 2020
#$twicNbr = "1312" # The week in chess last number with illegal castling in the PGN
#$twicNbr = "1296" # The week in chess last number containing Chess960 variant
$settingFullTwic = Join-Path -Path $dirTwic -ChildPath "Twic2Scid.psd1" # will be created with the last twicNbr

#$baseRef = "GiantBase" # your own Scid base of reference. Zip to backup your base of reference before running the script.
$baseRef = "RefTest" # in a first time, prefer any smaller Scid base for test before using your real giant base

# (*) =>>> SECURITY: review the functions: DeleteFile, DeleteBase, CleanTemporaryFile, DeleteChess960, BackupRefBase
$isWhatif = $false # $true means that deleteFile displays the command without deletion otherwise $false mean do the deletion
$isConfirm = $true # $true means deleteFile asks for confirmation otherwise $false means do not confirm
function DeleteFile([string]$filename)
{
    if (Test-Path -Path "$filename" -PathType leaf)
    {
        if ($isWhatif)
        {   Remove-Item $filename -WhatIf}      # trace the command with parameter without deletion
        else
        {
            if ($isConfirm)
            { Remove-Item $filename -Confirm} # ask for confirmation before the deletion
            else
            { Remove-Item $filename}          # deletion without confirmation
        }
    }
}

function TwicInit()
{   # Init twic number based variables $url, $fileDownloaded, $fileFullPgn prefixed at "script" scope
    $fileRoot = "twic$twicNbr" # local variable of this function unknown outside the function
    $fileZip = "$fileRoot" + "g.zip"
    $script:url = "https://www.theweekinchess.com/zips/$fileZip" # variable at script scope known outside the function

    $script:fileDownloaded = Join-Path -Path $dirTwic -ChildPath $fileZip
    $fileNamePgn = $fileRoot + ".pgn" # without path
    $script:fileFullPgn = Join-Path -Path $dirTwic -ChildPath $fileNamePgn #with full path
}

$rootBackup = "backup" # subfolder of the Scid bases to store a copy of the base of ref before its update
$dirBackup = Join-Path -Path $dirRef -ChildPath "$rootBackup" # folder will be created to save the base of ref
$baseFullRefWin = Join-Path -Path $dirRef -ChildPath $baseRef

$baseFullEcoWin = Join-Path -Path $pathScid -ChildPath (Join-Path -Path "data" -ChildPath "scid.eco") # Supplied by Scid
$baseFullEcoUnix = $baseFullEcoWin.Replace("\", "/") # Scid does not understand Windows single backslash but Unix slash

$ScidScript = Join-Path -Path $dirTwic -ChildPath "ScidScript.txt" # textual script file will be created
# Utility of type line of command in the Powershell console or Cmd console
$runPgnToBaseScid = Join-Path -Path $pathScid -ChildPath "pgnscid.exe" # Create a new Scid base from a PGN
$runScidShell = Join-Path -Path $pathScid -ChildPath "tcscid.exe"      # Interpreter to script a Scid base
$runScidCmd = Join-Path -Path $pathScid -ChildPath "scidt.exe"         # Scid base utility driven by options
# Windows tool requiring a mouse
$runScidGui = Join-Path -Path $pathScid -ChildPath "scid.exe"          # Graphic User Interface of Scid
$slotClipBase = 9 # 1. Run Scid. 2. menu "Help" > "Startup window". 3. above "Keep", enter in the input field the command: "sc_info clipbase" without double quotes
$slotFirstBase = 1 # First opened base
$runNotePad = "C:\Tool\NotepadPlus\Prog\notepad++.exe" # Must exist. Download from https://notepad-plus-plus.org/downloads/

# How to populate the template for IA-based ConvertFrom-String according to the programmation by examples?
# 1. Copy and paste the two most representative types of error from twic.err among:
# C:\Chess\base\twic\twic1312.pgn:(game 3270, line 86359) Warning: illegal castling (O-O) in game Gelfand, B - Gareev, T, 2019"
# C:\Chess\base\twic\twic1296.pgn:(game 2570, line 63828) Error: Invalid FEN: nrkrqbbn/pppppppp/8/8/8/8/PPPPPPPP/NRKRQBBN w DBdb - 0 1
# C:\Chess\base\twic\twic1296.pgn:(game 2570, line 63828) Error reading tag: [FEN "nrkrqbbn/pppppppp/8/8/8/8/PPPPPPPP/NRKRQBBN w DBdb - 0 1"]
# C:\Chess\base\twic\twic1296.pgn:(game 2570, line 63836) Error reading move in game Kasparov, G - Caruana, F, 2019: Nb3
# C:\Chess\base\twic\twic1296.pgn:(game 2571, line 63854) Error: Invalid FEN: nrkrqbbn/pppppppp/8/8/8/8/PPPPPPPP/NRKRQBBN w DBdb - 0 1
# C:\Chess\base\twic\twic1296.pgn:(game 2571, line 63854) Error reading tag: [FEN "nrkrqbbn/pppppppp/8/8/8/8/PPPPPPPP/NRKRQBBN w DBdb - 0 1"]

# 2. Let us choose line 80 and line 82 that represent the most typical and different error messages

# 3. Encapsultate each relevant string between curly brace "{" & "}" that the parser will extract keeping unchanged the existing delimiters ":", " ", "(", ")" and ","
# {C:\Chess\base\twic\twic1312.pgn}:(game {3270}, line {86359}) {Warning}: {illegal castling (O-O) in game Gelfand, B - Gareev, T, 2019}"
# {C:\Chess\base\twic\twic1296.pgn}:(game {2570}, line {63828}) {Error reading tag}: {[FEN "nrkrqbbn/pppppppp/8/8/8/8/PPPPPPPP/NRKRQBBN w DBdb - 0 1"]}
#                                  ^delimitor           ^ we are mainly interested in the line number to edit the PGN at this line of the error

# 4. Name as you want each encapsulated string by an unique identifier followed by ":" in the left part of each curly brace expression
# {pgn:C:\Chess\base\twic\twic1312.pgn}
#  ^ "pgn" has been chosen but "filename" or "path" could be the identifier of this first token. Each identifier is free.

# 5. Add a star "*" as suffix of the first key. In twic.err, the statement of one error is on one line.
# {pgn*:C:\Chess\base\twic\twic1312.pgn}

# 6. Use the here-string literal operator "@'" ... "'@" to define the group of the two samples of the template: ConvertFrom-String will learn it to generate internally the parser. 

$templateTwicErr = @'
{pgn*:C:\Chess\base\twic\twic1312.pgn}:(game {game:3270}, line {line:86359}) {type:Warning}: {error:illegal castling (O-O) in game Gelfand, B - Gareev, T, 2019}
{Pgn*:C:\Chess\base\twic\twic1296.pgn}:(game {game:2570}, line {line:63828}) {type:Error reading tag}: {error:[FEN "nrkrqbbn/pppppppp/8/8/8/8/PPPPPPPP/NRKRQBBN w DBdb - 0 1"]}
'@

# IA-based ConvertFrom-String will deduce that the type of error is a string expression including blank before the next ":".
# The final detailed "error" msg is after ":" until EOL. See CheckPgn for usage that only uses the line number.
# ConvertFrom-String is an interesting alternative with regard to using regular expr ou string utilities to extract the line number between ", line " and ")"

function Warning([string]$msg)
{   # Display $msg beginning by an error number then stop the script
    Write-Host $msg; exit
}

function CheckConfig()
{   # Check that the above variables are correctly configured
    if (!(Test-Path -Path $dirTwic)) {Warning "1000: $dirTwic does not exist. Configure dirTwic"}
    Set-Location -Path $dirTwic # the default folder is the folder of this script
    if (!(Test-Path -Path $pathScid)) {Warning "1010: $pathScid does not exist. Configure pathScid"}
    Write-Host "The base of reference to be updated is: $baseFullRefWin"
    Write-Host "You can press Ctrl+C to STOP the script at any time to keep the generated files."
    if (!(isExist($baseFullRefWin))) {Warning "1020: This base of reference does not exist. Configure baseRef, dirRef"}
    if (!(Test-Path -Path $baseFullEcoWin -PathType leaf)) {Warning "1030: $baseFullEcoWin does not exist. Configure baseFullEcoWin"}
    if (!(Test-Path -Path $runPgnToBaseScid -PathType leaf)) {Warning "1040: $runPgnToBaseScid does not exist. Configure runPgnToBaseScid"}
    if (!(Test-Path -Path $runScidShell -PathType leaf)) {Warning "1050: $runScidShell does not exist. Configure runScidShell"}
    if (!(Test-Path -Path $runScidCmd -PathType leaf)) {Warning "1060: $runScidCmd does not exist. Configure runScidCmd"}
    if (!(Test-Path -Path $runScidGui -PathType leaf)) {Warning "1070: $runScidGui does not exist. Configure runScidGui"}
    if (!(Test-Path -Path $runNotePad -PathType leaf)) {Warning "1080: $runNotePad does not exist. Configure runNotePad"}
    if (Test-Path -Path $settingFullTwic -PathType leaf)
    {
        $script:twicNbr = Get-Content -Path $settingFullTwic # retrieve the last twicNbr of the previous session
    }
}

function isAnswerYes([string]$question)
{   # $true if the user answers "y" otherwise $false in particular when the user presses the Enter key
    return ((Read-Host -Prompt "$question (y/[n])") -eq "y")
}

function DownloadPgn()
{   # Download the weekly PGN from the Internet as a zipped compressed archive .zip
    Write-Host "Try to download $fileDownloaded from $url"
    Write-Host "Please wait..."

    $start_time = Get-Date
    Invoke-WebRequest -Uri $url -OutFile $fileDownloaded
    Write-Host "Time: $((Get-Date).Subtract($start_time).Seconds)s"
}

function isChess960()
{   # $true is there is the tag "Chess960" in the PGN otherwise $false
    return Select-String -Path $fileFullPgn -SimpleMatch -CaseSensitive -Pattern "Chess960" -Quiet
}

function ScidBaseInfo([string]$msg, $ScidBase)
{   # Display in the console the stats about the generated Scid base of the week
    Write-Host "$msg"
    Out-String -InputObject (& "$runScidCmd" -i $ScidBase) | Write-Host
}

function RunScidScipt($scriptCmd)
{   # Generate the textual script file then run it in the Scid interpretor
    DeleteFile $ScidScript # delete previous Scid script
    $streamOut = [System.IO.StreamWriter] $ScidScript # generate dynamically the Scid script
    $scriptCmd | ForEach {$streamOut.WriteLine($_)} # populate the script file with each command
    $streamOut.WriteLine("sc_base close") # Close the twic database
    $streamOut.WriteLine("exit") # exit from the Scid interpreter
    $streamOut.Close()
    & "$runScidShell" $ScidScript    # invoke external exe    
}

function ScidCmdCompactBase()
{   # Array of Scid commands to compact the current base according to the filtered games to be deleted
    return  "sc_game flag D filter 1",          # set the "D"elete flag for all filtered games
            "sc_compact games",                 # Delete physically all filtered games
            "sc_compact names",                 # Removes all unused names if necessary
            "sc_filter reset"                   # Reset the current filter       
}

function ScidCmdClassifyEco()
{   # Array of Scid commands to classify the ECO code of all games of the current base
    return  "sc_eco read $baseFullEcoUnix",     # Read ECO codes in order to classify chess games
            "sc_eco base 1 1"                   # Reclassifies all games by ECO code with Scid extension
}

function DeleteChess960()
{   # Delete physically games with Chess960 variant in the generated scid database
    $scriptCmd =    "sc_base open $ScidTheWeekInChess", # Open the twic base of the week in Scid format
                    "sc_search header -pgn Chess960"    # Filter games containing the variant "Chess960"                    
    $scriptCmd += ScidCmdCompactBase; $scriptCmd += ScidCmdClassifyEco
    ScidBaseInfo "Before deleting Chess960:" $ScidTheWeekInChess
    RunScidScipt $scriptCmd
    ScidBaseInfo "After deleting Chess960 in the generated base but not in the PGN:" $ScidTheWeekInChess
    if (isAnswerYes("Do you want to REVIEW $ScidTheWeekInChess in the GUI of Scid"))
    {   Write-Host "REVIEW $ScidTheWeekInChess in the GUI of Scid"
        Start-Process -FilePath $runScidGui -ArgumentList "$ScidFullTwic" -WindowStyle Normal
        Write-Host
    }
}

function EditPgnAtLine([string]$filename, [string]$lineNbr)
{   # Run Notepad++ to edit $filename. Go automatically to $lineNbr where the issue to be fixed should be near this line number
    if (isAnswerYes("Do you want to EDIT $filename in NotePad++ at line $lineNbr"))
    {   
        if ($lineNbr -gt 1) {Set-Clipboard $lineNbr} # The line number is copied in the clipboard to be pasted in the dlgbox "Go To"
        Write-Host "Fix the PGN issue editing in NotePad++ $filename"
        $process = Start-Process -FilePath $runNotePad -ArgumentList "$filename" -WindowStyle Normal -PassThru
        if ($lineNbr -gt 1)
        {
            $null = $process.WaitForInputIdle() # Wait that NotePad++ GUI and the $filename are loaded
            [void][reflection.assembly]::loadwithpartialname("system.windows.forms")
            [system.windows.forms.sendkeys]::sendwait('^g') # Ctrl+G means NotePad++ menu 'Search' > 'Go To...'
            [system.windows.forms.sendkeys]::sendwait('^v') # Ctrl+V means paste the line number of the error            
            [system.windows.forms.sendkeys]::sendwait('{ENTER}') # Validate by pressing the ENTER key of the dialog box in NotePad++
        } # If NotePad++ does NOT go the wished line number, update the shortcuts according to the locale of your NotePad++ GUI
        Write-Host "Once the PGN fixed, RESTART this script answering 'n' or 'enter' about download and unzip"
    }
}

function CheckPgn()
{   # $true if PGN ok otherwise $false: display the PGN error in the console
	DeleteFile "$ScidTheWeekInChess.err" #delete previous error
    # create the twic database in Scid format from the unzipped PGN
    $sts = & "$runPgnToBaseScid" -f $fileFullPgn $ScidTheWeekInChess # invoke external exe
    Out-String -InputObject $sts -Stream | Write-Host  # Avoid in boolean function: Write-Output $sts
    if (Test-Path -Path "$ScidTheWeekInChess.err" -PathType leaf) # error in PGN?
    {   Write-Host
        $msg = Get-Content -Path "$ScidTheWeekInChess.err" -TotalCount 5 # first five errors only
        $msg | ConvertFrom-String -TemplateContent $templateTwicErr -OutVariable TwicErr | Out-Null # Parse error(s)
        if ($TwicErr.Count -gt 0) {$lineNbr = $TwicErr[0].line} else {$lineNbr = 0} # Copy the line number of the first error in the clipboard
        out-string -InputObject $msg | Write-Host # Display the first five errors of twic.err
        if (isChess960)
        {
            if (isAnswerYes("Do you want to REMOVE Chess960 games from the generated Scid $ScidTheWeekInChess"))
            {   DeleteChess960; if (isAnswerYes("Are all issues FIXED")) {return $true} 
            }
        }
        elseif (($TwicErr.Count -gt 1) -and (isAnswerYes("Do you want to READ $ScidFullTwic.err in NotePad++")))
        {   Write-Host "READ all issues in $ScidTheWeekInChess.err"
            Start-Process -FilePath $runNotePad -ArgumentList "$ScidFullTwic.err" -WindowStyle Normal
        }   
        EditPgnAtLine $fileFullPgn $lineNbr
        return $false   # Keep temporary files for debugging purpose
    }
    else {return $true} # The PGN is ok so we can merge the generated twic database in Scid format with the base of ref
}

function isExist([string]$ScidBase)
{   # #true if the three files of the $ScidBase really exists otherwise $false
    if (!(Test-Path -Path "$ScidBase.sg4" -PathType leaf)) {return $false}
    if (!(Test-Path -Path "$ScidBase.si4" -PathType leaf)) {return $false}
    if (Test-Path -Path "$ScidBase.sn4"   -PathType leaf) {return $true}
    return $false
}

function DeleteBase([string]$ScidBase)
{   # Delete physically the three files of the $ScidBase. Try with a base of test before using your real giant Scid base.
    DeleteFile "$ScidBase.sg4"
    DeleteFile "$ScidBase.si4"
    DeleteFile "$ScidBase.sn4"
}

function BackupRefBase()
{   # Backup .sg4, .si4, .sn4 of the Scid base of reference in subfolder "backup"
    if (Test-Path -Path $dirBackup -PathType Container)
    {
        $oldBackup = Join-Path -Path $dirBackup -ChildPath "$baseRef"
        DeleteBase $oldBackup
    }
    else
    {
        New-Item -Path $dirRef -Name $rootBackup -ItemType "directory" | Out-Null # Create subfolder
    }
    Copy-Item -Path "$baseFullRefWin.*" -Destination "$dirBackup"
    ScidBaseInfo "Save the Scid base of reference in $dirBackup" $baseFullRefWin
    Write-Output @(Get-ChildItem -Path $dirBackup -Filter "$baseRef.*" | select LastWriteTime, Length, Name) | Out-String | Write-Host
}

function CleanTemporaryFile()
{   # Delete generated files when Pgn is successfully imported in the base of ref
    DeleteFile $ScidScript         # delete generated Scid script
    DeleteFile $fileFullPgn        # delete the unzipped PGN
    DeleteBase $ScidTheWeekInChess # delete the temporary weekly database in Scid format
}   # you can later remove   -confirm if you do not want to confirm any more each deletion. See also DeleteBase

function ScidScriptUpdateRef()
{   # Generate and run the textual Scid script to optimize the Scid twic base, classify each ECO and update the base of ref
    $baseFullRefUnix = $baseFullRefWin.Replace("\", "/") # Scid does not understand Windows single backslash

    $optDup = "-colors 1 -event 1 -site 1 -year 1 -month 1 -day 1 -result 1 -eco 1 -moves 1 -comments 1 -variations 1 -setfilter 1"
    $scriptCmd =    "sc_base open $ScidTheWeekInChess", # Open the Scid base of the week
                    "sc_base duplicates $optDup"        # Set the filter of duplicated games

    # the clipbase is used as temporary base in memory in order to work with only one physical base in the first slot
    $ScidClipbase = "sc_filter copy $slotFirstBase $slotClipBase", # copy all games from the twic base to clipbase
                    "sc_base close",
                    "sc_base switch $slotClipBase",  # The clipbase becomes the active base    
                    "sc_base open $baseFullRefUnix", # Open the base of reference, which is the target of the update
                    "sc_filter copy $slotClipBase $slotFirstBase" # copy all games from the clipBase to the base of reference
    $scriptCmd += ScidCmdCompactBase; $scriptCmd += ScidCmdClassifyEco; $scriptCmd += $ScidClipbase
    RunScidScipt $scriptCmd
    ScidBaseInfo "After MERGING the new PGN in Scid base of reference: $baseFullRefUnix" $baseFullRefWin
    Write-Output @(Get-ChildItem -Path $dirRef -Filter "$baseRef.*" | select LastWriteTime, Length, Name) | Out-String | Write-Host
    CleanTemporaryFile
}

function ProcessTwic()
{   # $true if download weekly twic, unzip, check PGN OK, backup and update the base of ref otherwise $false 
    if (isAnswerYes("Do you want to DOWNLOAD $url")) {DownloadPgn}
    if (Test-Path -Path $fileDownloaded -PathType leaf)
    {
        if (isAnswerYes("Do you want to UNZIP .zip to generate $fileFullPgn"))
        {
            DeleteFile $fileFullPgn # delete alreay dezipped PGN
            Expand-Archive -Path $fileDownloaded -DestinationPath $dirTwic # Unzip the downloaded archive
            if (Test-Path -Path $fileFullPgn -PathType leaf)
            {   Write-Host "$fileFullPgn has been successfully unzipped." }
        } else {Write-Host "Process existing $fileDownloaded"}
        if (Test-Path -Path $fileFullPgn -PathType leaf)
        {   
            $isOk = CheckPgn # is the downloaded PGN understood by Scid?
            if ($isOk) 
            {   
                if (isAnswerYes("Do you want to UPDATE the base of reference $baseFullRefWin"))
                {
                    BackupRefBase
                    ScidScriptUpdateRef # to merge the downloaded PGN in the Scid base of reference
                    return $true # The base of reference has been successfully updated
                } else {Write-Host "Temporary files are kept for debugging purpose"}
            }
            elseif ($isOk -ne $false) {Warning "2000: CheckPgn returned: $isOk"}
        }   else {Warning "2010: $fileFullPgn is not found"}
    }   else {Warning "2020: $fileDownloaded is not downloaded"}
    return $false # There is at least one unfixed issue in the PGN or the generated Scid twic base
}

# main
CheckConfig
do
{
    TwicInit                     # Init twic number based variables $url, $fileDownloaded, $fileFullPgn
    if (!(ProcessTwic)) {exit}   # download, unzip, create twic.si4, compact, ECO classify then update the base of ref
    $twicNbr = [int]$twicNbr + 1 # increment the twicNbr in order to process the next week
    Set-Content -Path $settingFullTwic -Value $twicNbr # Persist on hard disk the next twic number
}   while (isAnswerYes("Do you want to process the NEXT twic $twicNbr"))
# end Twic2Scid.ps1
